;; The first two consecutive numbers to have two distinct prime factors
;; are:

;; 14 = 2 x 7
;; 15 = 3 x 5

;; The first three consecutive numbers to have three distinct prime
;; factors are:

;; 644 = 2^2 x 7 x 23
;; 645 = 3 x 5 x 43
;; 646 = 2 x 17 x 19.

;; Find the first four consecutive integers to have four distinct primes
;; factors. What is the first of these numbers?

(defmacro acollect (&body body)
  (let ((resv (gensym)))
    `(let ((it (list))
	   (,resv nil))
      (setf ,resv (progn ,@body))
      (values it ,resv))))

(declaim (inline factor))

(defun factor (num)
  (declare (optimize speed)
	   (fixnum num))
  (acollect
    (push 1 it)
    (loop for ff from 2 to (+ 1 (ash num -1)) ;; optimization, close enough to (ceiling (/ num 2))
	  do
	  (when (and (= 0 (mod num ff))
		     (not (member ff it :test #'=)))
	    (push ff it)))))



(defun memoize (function)
  (let ((res-table (make-hash-table :test #'equal :size 100000)))
    (lambda (&rest args)
      (multiple-value-bind 
	    (value exists) 
	  (gethash args res-table)
	(when (not exists)
	  (setf value (apply function args))
	  (setf (gethash args res-table) value))
	value))))

(defun inner-prime? (num)
  (cond ((= 1 num) nil)
	((= 2 num) t)
	((= 0 (mod num 2)) nil)
	(t
	 (loop for fact = (ceiling (sqrt num))
	       then (if (oddp fact)
			(- fact 2)
			(- fact 1))
	       while (>= fact 2)
	       do
	       (when (= 0 (mod num fact))
		 (return-from inner-prime? nil)))
	 t)))


(defvar *inner-prime?-mem* (memoize #'inner-prime?))

(defun prime? (num)
  (funcall *inner-prime?-mem* num))

(defun prime-factor (facts num)
  (cond ((= 1 num)
	 facts)
	((prime? num)
	 (cons num facts))
	(t
	 (loop for ii from 2 to (floor (/ num 2))
	       do
	       (when (= 0 (mod num ii))
		 (return-from prime-factor (prime-factor (cons ii facts) (/ num ii))))))))

;; (prime-factor (list) 14)
;; (prime-factor (list) 644)
;; (prime-factor (list) 645)
;; (prime-factor (list) 646)

(defun list-unique (things)
  (let ((ht (make-hash-table))
	(result (list)))
    (loop for thing in things
	  do
	  (when (not (gethash thing ht))
	    (setf (gethash thing ht) t)
	    (push thing result)))
    (reverse result)))

(defun has-n-distinct-prime-factors (num nn)
  (= nn (length (list-unique (prime-factor (list) num)))))

;;(has-n-distinct-prime-factors 644 3)

(time
 (loop for ii = 646 then (+ ii 1)
       ;;repeat 1000
       do
       (when (and (has-n-distinct-prime-factors ii 4)
		  (has-n-distinct-prime-factors (+ ii 1) 4)
		  (has-n-distinct-prime-factors (+ ii 2) 4)
		  (has-n-distinct-prime-factors (+ ii 3) 4))
	 (format t "found: ii:~a~&" ii)
	 (return ii))))
